
open "cartesian_h";
open "rel_h";
open "sort_h";
open "gensym_h";

(*
fun avars (ATOM (_, l)) = union {tvars t | t in list l};
*)

fun is_flat_term (V _) = NONE (* returns NONE if term is flat (variable or function
			       applied to variables), otherwise SOME (f, l)
			       where term is f $ l. *)
  | is_flat_term (f $ l) =
    if all
	(case t of
	     V _ => true
	   | _ => false)
       | t in list l
       end
	then NONE
    else SOME (f, l);

(*
fun tabbrv abbrv =
    let fun tab t =
	    if t inset abbrv
		then ?abbrv t
	    else case t of
		     V _ => t
		   | f $ l => f $ [tab t | t in list l]
    in
	tab
    end

fun aabbrv abbrv =
    let val tab = tabbrv abbrv
    in
	fn ATOM (P, l) => ATOM (P, [tabbrv t | t in list l])
    end

fun labbrv abbrv =
    let val aab = aabbrv abbrv
    in
	fn (sign, atom) => (sign, aab atom)
    end

fun gabbrv abbrv = map (labbrv abbrv);
*)

fun new_pred_list (P, l) =
    let val n = ref 0
	val f as |[put, convert, seek, tell, truncate, ...]| = outstring P
	val pos = (put "/"; tell ())
    in
	[
	  (inc n;
	   seek pos;
	   truncate ();
	   print f (pack (!n));
	   convert ())
	| _ in list l]
    end;

val sort_blocks = sort (fn ((m, _), (n, _)) => m<n);

fun compare_blocks ((pos, neg), (pos', neg')) =
    let val n = card pos and n'=card pos'
    in
	n<n' orelse
	n=n' andalso card neg < card neg'
    end;
val sort_simple_blocks = sort (fn (C_BLOCK b, C_BLOCK b') => compare_blocks (b, b'));

fun compare_block_lists (nil, nil) = false
  | compare_block_lists (nil, _) = true
  | compare_block_lists (_, nil) = false
  | compare_block_lists ((i, block)::rest, (i', block')::rest') =
    i<i' orelse
    i=i' andalso
    (compare_blocks (block, block')
     orelse block=block' andalso
     compare_block_lists (rest, rest'));

fun compare_complexes (C_COMPLEX ((pos, neg), f, eqvars, blocks),
		       C_COMPLEX ((pos', neg'), f', eqvars', blocks')) =
    let val n = card eqvars and n' = card eqvars'
    in
	n<n' orelse
	n=n' andalso
	let val m = card pos and m' = card pos'
	in
	    m<m' orelse
	    m=m' andalso
	    let val p = card neg and p' = card neg'
	    in
		p<p' orelse
		p=p' andalso
		compare_block_lists (blocks, blocks')
	    end
	end
    end;
val sort_complex_blocks = sort compare_complexes;

(* get_components (pre-gclause, simple, acc)
 approximates a pre-gclause which is a disjunction of literals of
 the form +-P(x) or +-P(f(x1,...,xk)) by:
 - accumulating all +-P(x) in simple, which maps every variables x
 to the pair (pos, neg),
 pos=set of P with +P(x) present, neg=set of P with -P(x) present;
 - accumulating all +-P(f(x1,...,xk)) in acc, which maps every term f(x1,...,xk)
 to the pair (pos, neg),
 pos=set of P with +P(f(x1..xk)) present, neg=set of P with -P(f(x1..xk)) present.
 Then it produces a disjunction of clauses:
 - for each term f(x1..xk), the disjunction of +-P(f(x1..xk)) in acc,
 in disjunction with all +-P(x) in simple such that x is one of x1,...,xk;
 (this is gathered in the map complexes below;)
 - for each variable x not free in any f(x1..xk) in acc,
 the block of all +-P(x) in simple.
 *)
fun get_components (nil, simple, acc) =
    let val simpler = ref simple
	val complexes =
	    {let val tnums = ref {}
		 val tnum = ref 0
		 val varsinv = {(inc tnum; !tnum) =>
				    if t inset !tnums
					then ?(!tnums) t
				    else (tnums := !tnums ++ {t => !tnum};
					  !tnum)
			       | t in list l}
		 val eqvars = invrel varsinv
		 val blocks = sort_blocks {(simpler := {x} <-| !simpler;
					    (?(!tnums) t, ?simple x))
					  | t as V x in list l
					      such that x inset simple}
	     in
		 C_COMPLEX (block, f, eqvars, blocks)
	     end
	    | (f $ l) => block in map acc}
	val simples = sort_simple_blocks {C_BLOCK b
					 | _ => b in map !simpler}
    in
	simples @ sort_complex_blocks complexes
    end
  | get_components ((sign, P, t)::rest, simple, acc) =
    case t of
	V x => let val simple' =
		       if x inset simple
			   then simple ++ {x => let val (pos, neg) = ?simple x
						in
						    if sign
							then (pos U {P}, neg)
						    else (pos, neg U {P})
						end}
		       else simple ++ {x => if sign then ({P}, {}) else ({}, {P})}
	       in
		   get_components (rest, simple', acc)
	       end
      | _ => get_components (rest, simple, acc ++ {t => if t inset acc
							    then let val (pos, neg) = ?acc t
								 in
								     if sign
									 then (pos U {P}, neg)
								     else (pos, neg U {P})
								 end
							else if sign
							    then ({P}, {})
							else ({}, {P})});

fun add_cart_approx_clause (fsig, predsig, tableau, GCLAUSE (name, _, pregc)) =
    let val clauses = ref nil
	val fsigr = ref fsig
	val psigr = ref predsig
	fun cart_approx_clause (name, nil, acc) =
	    clauses := (get_components (acc, {}, {}), J_AX (ID_NAME name)) :: !clauses
	  | cart_approx_clause (name, (lit as (sign, ATOM (P, l)))::rest, acc) =
	    case l of
		[t] =>
		  (case is_flat_term t of
		       NONE => cart_approx_clause (name, rest, (sign, P, t)::acc)
		     | SOME (f, l) => 
		       (* original clause is +-P(f(t1,...,tn)) or rest or acc
			(where l = [t1,...,tn]).
			Split it into:
			. -+P1(x1) or +-P(f(x1,...,xn))
			...
			. -+Pn(xn) or +-P(f(x1,...,xn))
			. +-P1(t1) or ... or +-Pn(tn) or rest or acc
			where P1...Pn are fresh predicates
			(printed P/1, ..., P/n: in list Pl)
			x1...xn are fresh variables (for more precision,
			if ti=tj we let xi=xj, and we replace any term
			in rest or acc of the form ti by the corresponding xi).
			*)
		       let val abbrv = ref {}
			   val newpos = [(case t of
					      V _ => false
					    | _ =>
					      if t inset !abbrv
						  then false
					      else (abbrv := !abbrv ++ {t => V (gensym "_X")};
						    true)
						  )
					| t in list l]
			   val Pl = new_pred_list (P, l)
			   val tnums = ref {}
			   val tnum = ref 0
			   val varsinv = {(inc tnum; !tnum) =>
					      if t inset !tnums
						  then ?(!tnums) t
					      else (tnums := !tnums ++ {t => !tnum};
						    !tnum)
					 | t in list l}
			   val eqvars = invrel varsinv
			   val pblock = if sign then ({P}, {}) else ({}, {P})
		       in
			   tnum := 0;
			   (* create clauses -+Pi(xi) or +-P(f(x1,...,xn)) *)
			   iterate
			     (inc tnum;
			      if newp
				  then clauses :=
				      ([C_COMPLEX (pblock, f, eqvars,
						   [(!tnum,
						     if sign
							 then ({}, {P'})
						     else ({P'}, {}))])],
					 J_AX (ID_NAME name))
					 :: !clauses
			      else ())
			   || t in list l and newp in list newpos and P' in list Pl
			   end;
		       (* now for clause +-P1(t1) or ... or +-Pn(tn) or rest or acc:
			(we don't include Pi(ti) and Pj(tj) when ti=tj).
			*)
			cart_approx_clause (name,
					    [(sign, ATOM (P', [t]))
					    || t in list l and true in list newpos
					    and P' in list Pl] @ rest,
					      acc)
		       end)
	      | _ => (* in particular P() is coded as P (_f_P ()) *)
		let val f = "_f_" ^ P
		    val Psort = ?(!psigr) P handle MapGet => raise NoPredSig P
		    val typeP = "type-" ^ P
		in
		    fsigr := !fsigr ++ {f => ("o", Psort)};
		    psigr := !psigr ++ {typeP => ["o"]};
		    cart_approx_clause (name, (sign, ATOM (typeP, [f $ l]))::rest, acc)
		end
    in
	cart_approx_clause (name, pregc, nil);
	(add_clause_lists (tableau, !clauses), !fsigr, !psigr)
    end;
