
open "cartesian_h"
open "inline_h"
open "magic_h"
open "partition_h"
open "engine_h"

exception UnknownTransformer of transformer

fun order_of_prec prec =
    let val matchlist = [#matches (regexp re)
			| re in list prec]
	fun rank1 (P, n, nil) = n
	  | rank1 (P, n, matches::l) =
	    if matches P
		then n
	    else rank1 (P, n+1, l)
	memofun rank P = rank1 (P, 0, matchlist)
	fun compare (P, Q) =
	    let val m = rank P and n = rank Q
	    in
		m>n orelse
		m=n andalso P strless Q
	    end
    in
	compare
    end;

fun print_transformer (f as |[put, ...]|) =
    let fun pt (T_INLINE s) = (put " -inline ";
			     put s)
	  | pt T_MAGIC = put " -magic"
	  | pt (T_PARTITION (pred_re, sort_re)) =
	    (put " -partition ";
	     put pred_re;
	     put " ";
	     put sort_re)
	  | pt t = raise UnknownTransformer t
    in
	pt
    end;

fun engine (fsig, prog, goals, trans, less, verbose) =
    let val progr = ref prog
	fun verb do_f =
	    if verbose
		then do_f ()
	    else ()
	val ptrans = print_transformer stderr
    in
	verb (fn () =>
		 (#put stderr "Starting engine.  Initial clause set:\n";
		  print_program stderr prog;
		  #flush stderr ()));
	iterate
	  (verb (fn () =>
		    (#put stderr "\n*";
		     ptrans tr;
		     #put stderr "... ";
		     #flush stderr ()));
	   case tr of
	       T_INLINE pred_re => progr := inline_predicates pred_re (!progr)
	     | T_MAGIC => progr := magic_templates (!progr, goals)
	     | T_PARTITION re => progr := partition_clauses re (fsig, !progr)
	     | _ => raise UnknownTransformer tr;
	       verb (fn () =>
			(#put stderr "done.  Current clause set:\n";
			 print_program stderr (!progr))))
	| tr in list trans
	end;
	verb (fn () =>
		 (#put stderr "* Converting to automaton by cartesian approximation... ";
		  #flush stderr ()));
	let val PROG (predsig, pr) = !progr
	    val fsigr = ref fsig
	    val psigr = ref predsig
	    val tableau = ref tableau_true
	    fun add cl =
		let val (tab', fsig', psig') =
			add_cart_approx_clause (!fsigr, !psigr,
						!tableau, cl)
		in
		    tableau := tab';
		    fsigr := fsig';
		    psigr := psig'
		end
	in
	    iterate
	      iterate
		add (gclause_of_definite dcl)
	      | dcl in list dcls
	      end
	    | P => dcls in map pr
	    end;
	    verb (fn () =>
		     (#put stderr "adding goals";
		      iterate
			(#put stderr " ";
			 #put stderr g)
		      | g in list goals
		      end;
		      #put stderr "... ";
		      #flush stderr ()));
	    iterate
	      (psigr := !psigr ++ {g => nil};
	       add (GCLAUSE ("goal-" ^ g, {}, [(false, ATOM (g, nil))])))
	    | g in list goals
	    end;
	    verb (fn () => (#put stderr "done:\n";
			    print_tableau ({f => len l
					   | f => (_, l) in map !fsigr},
					     stderr) (!tableau);
			    #put stderr "* Saturating... ";
			    #flush stderr ()));
	    expand_tableau less (!tableau)
	    before verb (fn () =>
			    (#put stderr "done.\n";
			     #flush stderr ()))
	end
    end;
