
open "funsig_h"
open "term_h"

datatype atom = ATOM of string * term list;

extern val avars : atom -> string set;

type pre_gclause = (bool * atom) list;

datatype gclause = GCLAUSE of string * sortenv * pre_gclause;
	 (* clause name, typing context, pre-gclause. *)

datatype definite_clause = DCLAUSE of string (* clause name *)
	 * sortenv (* typing context *)
	 * atom (* head *)
	 * atom list (* body *)
	 ;

extern val gclause_of_definite : definite_clause -> gclause;

exception NonDefinite;
extern val definite_of_gclause : gclause -> definite_clause;

datatype program = PROG of predsig
	 * (string -m> definite_clause list);
	 (* indexes clauses by predicate of head. *)

extern val print_atom : 'a outstream -> atom -> unit;
extern val print_definite_clause : 'a outstream -> definite_clause -> unit;
extern val print_program : |[put : string -> unit, flush : unit -> unit,
			      ... : 'a]| -> program -> unit;

extern val add_gclause : (string -m> definite_clause list)
	   * definite_clause -> string -m> definite_clause list;

extern val atom_subst : (string -m> term) -> atom -> atom;
extern val definite_clause_subst : (string -m> term) ->
	   sortenv * definite_clause -> definite_clause;

type goal = string (* name of 0-ary goal; we don't have more general goals,
		    since a query
		    ?P1(t1),...,Pn(tn)
		    which is the goal
		    false :- P1(t1), ..., Pn(tn)
		    is coded as two clauses:
		    signal :- P1(t1), ..., Pn(tn)
		    false :- signal. *)
