
(* A model is an up-tree automaton, i.e., a set
 of clauses of the form:
 (1) P1(x1), ..., Pn(xn) => P(f(x1,...,xn))  [where x1,...,xn are not necessarily distinct]
 (2) P(x)

 This is exactly what we get as a model from ordered resolution with splitting
 on each saturated non-closed branch in the case of Horn clauses.

 Proof: let B be a saturated set of clauses.  Saturated means:
 if C \/ +A and C' \/ -A' in B have an ordered resolvent C'', then
 C'' splits as a disjunction of blocks, one of which is in B, or
 is subsumed by one in B.
 In general, the ordered resolvent C'' is subsumed by some clause in B.

 Now take any smallest ordered resolution refutation pi from B + non P(t).
 If pi contains an ordered resolvent between two clauses in B, then the
 resolvent is subsumed by some clause C in B: replace this inference by
 the clause C itself.  Propagate downwards, replacing clauses by subsuming
 clauses in general (as in my book).  This produces a refutation again, which
 is smaller: contradiction.

 So pi only resolves clauses from B with clauses not from B, or two clauses
 not from B.  Initially, the only possibility is to resolve non P(t) with
 some clause from B.  It is easy to see that the only clauses not from B that
 we may get are negative clauses.  They can only be resolved with clauses
 whose maximal literal is on the right.  This is the case for clauses of the
 form (1) or (2).  The other possible formats of clauses are:
   (3) Q1(x), ..., Qm(x) => Q(x)
or (4) B1(x1), ..., Bn(xn), Q1(f(x1..xn)), ..., Qm(f(x1..xn)) => Q(xi)
or (5) B1(x1), ..., Bn(xn), Q1(f(x1..xn)), ..., Qm(f(x1..xn)) => Q(f(x1..xn))
 where B1, ..., Bn are blocks (possibly empty).

 Now if ordered resolution is taken so that from each same-weight literal,
 negative literals are preferred (what I called negatively biased resolution,
 which is an instance of ordered resolution with selection), the greatest
 literal L is on the left, so none of these clauses can actually be used.  QED.
 *)
