
open "clauses";

val expand = expand_tableau (op strless);

val c_P_a_Q = C_COMPLEX (({"Q"}, {}), "a", {}, [(1, ({}, {"P"}))]);
val c_Q_b_P = C_COMPLEX (({"P"}, {}), "b", {}, [(1, ({}, {"Q"}))]);
val c_P_final = C_COMPLEX (({"P"}, {}), "epsilon", {}, []);

(* to test whether state P of S is empty, we may do:
 - if S,-P(x) |- false, then S |- exists x.P(x), so S (least model) is not empty.
 (i.e., S is not empty iff tableau reduces to [].)
 *)

val q_P = C_BLOCK ({}, {"P"});

val c = expand [BRANCH |[clauses={}, present={}, pos_index={}, neg_index={},
			  simple_queue = [ (c_P_a_Q, J_AX ID_NONE),
					   (c_Q_b_P, J_AX ID_NONE),
					   (c_P_final, J_AX ID_NONE),
					   (q_P, J_AX ID_NONE)],
			  complex_queue = nil,
			  deductions = {}]|];

(* c = [] *)

val csigt = {"c" => 2, "t" => 1, "nil" => 0, "cons" => 2, "d" => 1,
	      "k" => 1, "symk" => 1, "asymk1" => 1, "asymk2" => 1};

fun complete_knows (oldknowP, newknowP, tab) =
    let val knowlist = newknowP ^ "-list" (* "we know every element of the list" *)
	val knowinv = newknowP ^ "-inv" (* "we know the inverse of the argument" *)
	val knowkey = newknowP ^ "-key" (* "we know this key", of type key *)
	val knowkeyinv = newknowP ^ "-key-inv" (* "we know the inverse of this key" *)
	val knowkey_asymk1 = newknowP ^ "-asymk1" (* "we know the asymk1 of this key" *)
	val knowkey_asymk2 = newknowP ^ "-asymk2" (* "we know the asymk2 of this key" *)
	val newclauses = [ (*** I. Axiom ***)
			   (* oldknowP(x) => newknowP(x) *)
			   (C_BLOCK ({newknowP}, {oldknowP}), J_AX (ID_NAME (oldknowP ^ " -> " ^ newknowP))),

			   (*** II. Introduction rules. ***)
			   (* newknowP(x), newknowP(y) => newknowP(c (x,y)) *)
			   (C_COMPLEX (({newknowP}, {}), "c", {}, [(1, ({},{newknowP})), (2, ({},{newknowP}))]),
			    J_AX (ID_NAME (newknowP ^ " crypt I"))),
			   (* newknowP-list(x) => newknowP (t (x)) *)
			   (C_COMPLEX (({newknowP}, {}), "t", {}, [(1, ({},{knowlist}))]),
			    J_AX (ID_NAME (newknowP ^ " tuple I"))),
			   (* newknowP-list (nil) *)
			   (C_COMPLEX (({knowlist}, {}), "nil", {}, nil),
			    J_AX (ID_NAME (newknowP ^ " nil I"))),
			   (* newknowP(x), newknowP-list(y) => newknowP-list (cons (x, y)) *)
			   (C_COMPLEX (({knowlist}, {}), "cons", {}, [(1, ({},{newknowP})), (2, ({},{knowlist}))]),
			    J_AX (ID_NAME (newknowP ^ " cons I"))),

			   (*** III. Elimination rules. ***)
			   (* newknowP(c (x,y)), newknowP-inv (y) => newknowP(x) *)
			   (C_COMPLEX (({}, {newknowP}), "c", {}, [(1, ({newknowP},{})), (2, ({}, {knowinv}))]),
			    J_AX (ID_NAME (newknowP ^ " crypt E"))),
			   (* newknowP(t (x)) => newknowP-list(x) *)
			   (C_COMPLEX (({}, {newknowP}), "t", {}, [(1, ({knowlist},{}))]),
			    J_AX (ID_NAME (newknowP ^ " tuple E"))),
			   (* newknowP-list(cons(x,y)) => newknowP(x) *)
			   (C_COMPLEX (({}, {knowlist}), "cons", {}, [(1, ({newknowP},{}))]),
			    J_AX (ID_NAME (newknowP ^ " cons E1"))),
			   (* newknowP-list(cons(x,y)) => newknowP-list(y) *)
			   (C_COMPLEX (({}, {knowlist}), "cons", {}, [(2, ({knowlist},{}))]),
			    J_AX (ID_NAME (newknowP ^ " cons E2"))),

			   (*** IV. Knowing the inverse of a message *)
			   (* newknowP(d(x)) => newknowP-inv(d(x)) *)
			   (C_COMPLEX (({knowinv}, {newknowP}), "d", {}, nil),
			    J_AX (ID_NAME (newknowP ^ " inv data"))),
			   (* newknowP(c(x,y)) => newknowP-inv(c(x,y)) *)
			   (C_COMPLEX (({knowinv}, {newknowP}), "c", {}, nil),
			    J_AX (ID_NAME (newknowP ^ " inv crypt"))),
			   (* newknowP(t(x)) => newknowP-inv(t(x)) *)
			   (C_COMPLEX (({knowinv}, {newknowP}), "t", {}, nil),
			    J_AX (ID_NAME (newknowP ^ " inv tuple"))),
			   (* newknowP(k(x)) => newknowP-key(x) *)
			   (C_COMPLEX (({}, {newknowP}), "k", {}, [(1, ({knowkey},{}))]),
			    J_AX (ID_NAME (newknowP ^ " key I"))),
			   (* newknowP-key-inv(x) => newknowP-inv(k(x)) *)
			   (C_COMPLEX (({knowinv}, {}), "k", {}, [(1, ({}, {knowkeyinv}))]),
			    J_AX (ID_NAME (newknowP ^ " key inv E"))),
			   (* newknowP-key(symk(x)) => newknowP-key-inv(symk(x)) *)
			   (C_COMPLEX (({knowkeyinv}, {knowkey}), "symk", {}, nil),
			    J_AX (ID_NAME (newknowP ^ " key inv sym1"))),
			   (* newknowP-key(asymk1(x)) => newknowP-asymk1(x) *)
			   (C_COMPLEX (({}, {knowkey}), "asymk1", {}, [(1, ({knowkey_asymk1},{}))]),
			    J_AX (ID_NAME (newknowP ^ " key asym1"))),
			   (* newknowP-asymk1(x) => newknowP-key-inv(asymk2(x)) *)
			   (C_COMPLEX (({knowkeyinv}, {}), "asymk2", {}, [(1, ({},{knowkey_asymk1}))]),
			    J_AX (ID_NAME (newknowP ^ " key inv asym1"))),
			   (* newknowP-key(asymk2(x)) => newknowP-asymk2(x) *)
			   (C_COMPLEX (({}, {knowkey}), "asymk2", {}, [(1, ({knowkey_asymk2},{}))]),
			    J_AX (ID_NAME (newknowP ^ " key asym2"))),
			   (* newknowP-asymk2(x) => newknowP-key-inv(asymk1(x)) *)
			   (C_COMPLEX (({knowkeyinv}, {}), "asymk1", {}, [(1, ({},{knowkey_asymk2}))]),
			    J_AX (ID_NAME (newknowP ^ " key inv asym2")))
			   ]
    in
	expand (add_clauses (tab, newclauses))
    end;

val tab = complete_knows ("knows0", "knows1", tableau_true);

(*
 print_tableau (csigt, stdout) tab;
*)

val tab0 = expand (add_clauses (tableau_true,
				[ (C_BLOCK ({}, {"is-nonce-1", "is-nonce-2"}), J_AX (ID_NAME "nonce-1<>nonce-2")),
				  (C_COMPLEX (({"is-nonce-2"}, {}), "N2", {}, nil), J_AX (ID_NAME "exists nonce-2")),
				  (C_COMPLEX (({"is-nonce-1"}, {}), "N1", {}, nil), J_AX (ID_NAME "exists nonce-1"))
				  ]));

val tab5 = expand (add_clauses (tableau_true,
				[
				  (* knows0(x) => knows1(x) *)
				  (C_BLOCK ({"knows1"}, {"knows0"}), J_AX (ID_NAME "knows0 -> knows1")),
				  (* knows1(k(x)) => knows1-key(x) *)
				  (C_COMPLEX (({}, {"knows1"}), "k", {}, [(1, ({"knows1-key"},{}))]),
				   J_AX (ID_NAME "knows1 key I"))
				  ]));

	     
(*
val tab1 = expand (add_clauses (tableau_true,
				[ (C_BLOCK ({"knows0"}, {"is-nonce-1"}), J_AX (ID_NAME "knows0 nonce-1")), (* nonce 1 is known. *)
				  (C_BLOCK ({"knows0"}, {"is-nonce-2"}), J_AX (ID_NAME "knows0 nonce-2")), (* nonce 2 is known. *)
				  (C_BLOCK ({}, {"is-nonce-1", "is-nonce-2"}), J_AX (ID_NAME "nonce-1<>nonce-2")) (* but nonce 1 is not nonce 2. *)
				  ]));

val tab2 = complete_knows ("knows0", "knows1", tab1);
val tab3 = expand (add_clauses (tab2,
				[ (C_COMPLEX (({"is-M"}, {}), "c", {},
						  [(1,({},{"is-nonce-1"})), (2,({},{"is-nonce-2"}))]),
				   J_AX (ID_NAME "M:={nonce-1}_nonce_2")),
				  (* M is {nonce-1}_nonce-2. *)
				  (C_BLOCK ({}, {"is-M", "knows1"}), J_AX (ID_NAME "query knows1 M?"))
				  (* is M necessarily unknown? *)
				  ]));
val tab4 = expand (add_clauses (tab3,
				[ (C_COMPLEX (({"is-nonce-1"}, {}), "N1", {}, nil), J_AX (ID_NAME "N1 is nonce-1")),
				  (C_COMPLEX (({"is-nonce-2"}, {}), "N2", {}, nil), J_AX (ID_NAME "N2 is nonce-2"))
				  ]));
    *)

