%{
open Term
open List

let parse_error msg =
    (prerr_string msg;
     prerr_string " at characters ";
     prerr_int (symbol_start ());
     prerr_char '-';
     prerr_int (symbol_end ());
     prerr_newline ();
     exit 1);;

let type_prod pos tlst = match tlst with
  [] -> LINE (pos, APP ("unit", []))
| [t] -> t
| _ -> LINE (pos, APP ("prod", rev tlst));;

let say_know pos id t = LINE (pos, APP ("know", [id; t]));;

%}

%token <int> INT
%token <string> ID
%token <string> MSG_LABEL
%token <string> KW_CHAR
%token KW_HASH KW_ALIAS KW_KNOWS KW_INTRUDER KW_AXIOM KW_SWITCH KW_CASE
%token KW_LPAREN KW_RPAREN KW_LANGLE KW_RANGLE KW_EQUAL KW_COMMA KW_TST_EQUAL KW_ASSIGN
%token KW_PERCENT KW_COLON KW_LBRACE KW_RBRACE KW_ARROW KW_SUB KW_SUP KW_SESSION KW_EOF

%start spec
%type <Term.term> spec type terme bloc atomic_terme crypt_terme opt_sup session affectation
%type <term> propriete
%type <term list> declarations declaration formal_liste formal_liste_non_vide
%type <term list> tuple terme_liste environnement proprietes
%type <term list> affectations
%type <bool> hash_decl
%type <string> equal_op
%type <(term -> term)> messages message case
%type <(term -> term) list> cases

%%

spec : ID declarations bloc environnement proprietes
       KW_EOF { LINE ((symbol_start (), symbol_end ()),
                      APP ("spec",
	                   [LINE ((rhs_start 2, rhs_end 2), APP ("declare", rev $2));
                            LINE ((rhs_start 3, rhs_end 3),  $3);
                            LINE ((rhs_start 4, rhs_end 4),  APP ("sessions", rev $4));
                            LINE ((rhs_start 5, rhs_end 5),  APP ("props", rev $5))
		            ])) }
;

/* declarations est une liste inversee de declarations. */
declarations : { [] }
| declarations declaration { rev_append $2 $1 }
;

declaration : ids KW_COLON type { let pos = (symbol_start (), symbol_end ()) in
                                  map (fun id -> LINE (pos, APP ("type", [id; $3]))) $1 }
| ID KW_LPAREN formal_liste KW_RPAREN KW_COLON type hash_decl
  { [LINE ((symbol_start (), symbol_end ()),
          APP ("type",
               [VAR $1;
                LINE ((rhs_start 2, symbol_end ()),
                      APP ((if $7 then "one-way-function" else "function"),
                           [type_prod (symbol_start (), symbol_end ()) $3; $6]))]))] }
| KW_ALIAS ID KW_EQUAL terme
  { [LINE ((symbol_start (), symbol_end ()), APP ("alias", [VAR $2; $4]))] }
| ID KW_KNOWS tuple { map (say_know (symbol_start (), symbol_end ()) (VAR $1))
                                       (rev $3) }
| KW_INTRUDER KW_KNOWS tuple { map (say_know (symbol_start (), symbol_end ())
                                     (APP ("intruder", []))) (rev $3) }
| KW_AXIOM terme KW_EQUAL terme
  { [LINE ((symbol_start (), symbol_end ()), APP ("axiom", [$2; $4]))] }
;

/* ids est une liste inversee d'identificateurs. */
ids : ID { [ VAR $1 ] }
| ids KW_COMMA ID { (VAR $3)::$1}
;

type : ID { VAR $1 }
;

/* formal_liste est une liste inversee de types. */
formal_liste : { [] }
| formal_liste_non_vide { $1 }
;

/* formal_liste_non_vide est une liste inversee de types. */
formal_liste_non_vide : type { [$1] }
| formal_liste_non_vide KW_COMMA type { $3 :: $1 }
;

hash_decl : KW_HASH { true }
| { false }
;

bloc : KW_LBRACE messages KW_RBRACE { $2 (APP ("stop", [])) }
;

messages : { fun k -> k }
| messages message { fun k -> $1 ($2 k) }
;

message : MSG_LABEL ID KW_ARROW ID KW_COLON terme
  { let pos = (symbol_start (), symbol_end ()) in
    fun k -> LINE (pos, APP ("comm", [k; VAR $1; VAR $2; VAR $4; $6])) }
| ID KW_COLON ID equal_op terme
  { let pos = (symbol_start (), symbol_end ()) in
    fun k -> LINE (pos, APP ($4, [k; VAR $1; VAR $3; $5])) }
| KW_LBRACE messages KW_RBRACE { $2 }
| KW_SWITCH terme KW_LBRACE cases KW_RBRACE
  { let pos = (symbol_start (), symbol_end ()) in
    fun k -> LINE (pos, APP ("switch", $2::map (fun case -> case k) $4)) }
;

equal_op : KW_TST_EQUAL { "==" }
| KW_ASSIGN { ":=" }
;

/* cases est une liste inversee de cas. */
cases : { [] }
| cases case { $2::$1 }
;

case : KW_CASE terme KW_COLON messages
  { let pos = (symbol_start (), symbol_end ()) in
    fun k -> LINE (pos, APP ("case", [$2; $4 k])) }
;

terme : tuple { match $1 with
                  [] -> APP ("tuple", []) (* ne peut pas arriver *)
                | [t] -> t
                | tlst -> LINE ((symbol_start (), symbol_end ()),
                                APP ("tuple", rev tlst)) }
;

/* tuple est une liste inversee de termes. */
tuple : crypt_terme { [$1] }
| tuple KW_COMMA crypt_terme { $3::$1 }
;

/* terme_liste est une liste inversee de termes. */
terme_liste : KW_LPAREN KW_RPAREN { [] }
| KW_LPAREN tuple KW_RPAREN { $2 }
;

crypt_terme : atomic_terme { $1 }
| KW_LBRACE terme KW_RBRACE KW_SUB atomic_terme opt_sup
  { LINE ((symbol_start (), symbol_end ()),
          APP ("crypt", [$2; $5; $6])) }
| atomic_terme KW_PERCENT atomic_terme
  { LINE ((symbol_start (), symbol_end ()),
          APP ("percent", [$1; $3])) }
;

atomic_terme : ID terme_liste
  { LINE ((symbol_start (), symbol_end ()),
          APP ("app", (VAR $1)::(rev $2))) }
| ID { VAR $1 }
| KW_LPAREN terme KW_RPAREN { $2 }
| KW_LPAREN KW_RPAREN { LINE ((symbol_start (), symbol_end ()),
                              APP ("tuple", [])) }
;

opt_sup : { APP ("vanilla", []) }
| KW_SUP atomic_terme { $2 }
;

/* environnement est une liste inversee de sessions. */
environnement : { [] }
| environnement session { $2::$1 }
;

session : KW_SESSION affectations { LINE ((symbol_start (), symbol_end ()),
                              APP ("session", rev $2)) }
;

/* affectations est une liste inversee d'affectations de noms de principaux a des roles. */
affectations : affectation { [$1] }
| affectations KW_COMMA affectation { $3::$1 }
;

affectation : ID KW_EQUAL crypt_terme { LINE ((symbol_start (), symbol_end ()),
                              APP ("role", [VAR $1; $3])) }
;

/* proprietes est une liste inversee de proprietes a verifier. */
proprietes : { [] }
| proprietes propriete { $2::$1 }
;

propriete : ID terme_liste { LINE ((symbol_start (), symbol_end ()),
                              APP ("claim", VAR $1::rev $2)) }
;
